<?php
/**
 *
 * @license http://www.browsercrm.com/licence.php  BrowserCRM Software Licence
 * @copyright Copyright (c) 2003 - 2007 BrowserCRM Limited
 * @link http://www.browsercrm.com/
 * @package Services
 */

/**
 * FileSystem - this class handles the file IO functions for BrowserCRM
 *
 * @author Matthew Ross, Matthew Corich, Phil Calder, Brenton LeTrabb, Minggang Chen
 * @package Services
 */
class FileSystem extends Service {
	/**
	 * @var string the current directory you are working in
	 */
	var $directory = false;
	/**
	 * @var string the current error message
	 */
	var $error = false;
	/**
	 * @var string the desired chmod param
	 */
	var $chmod_string = 0770;
	
	/**
	 * @var int max files per (sub)directory
	 */
	var $max_files_per_dir = 1000;
	
	/**
	 * Change the directory you are working with
	 * 
	 * @param string $directory name
	 */
	function setDirectory($directory, $create = false) {
		global $USER, $L;
		if (@is_dir($directory)) {
			$this->directory = $directory;
			return true;
		} else {
			if ($create) {
				$create_result = $this->createDir(basename($directory), dirname($directory), true);
				if (!$create_result) {
					$this->setError($L['FileSystemErrorSetDirectoryFailed'] . ': ' . $directory );
				}
			} else {
				$this->setError($L['FileSystemErrorSetDirectoryFailed'] . ': ' . $directory );
			}
		}
		return false;
	}
	
	/**
	 * returns the directory you are working with
	 * 
	 * @return string $directory path
	 */
	function getDirectory() {
		return $this->directory;
	}
	
	/**
	 * set the chmod param
	 */
	function setChmodString($chmod_string) {
		$this->chmod_string = $chmod_string;
	}
	
	function getChmodString() {
		return $this->chmod_string;
	}
	
	function doChmod($path) {
		global $L;
		
		$old = umask(0);
		$result = @chmod($path, $this->getChmodString());
		umask($old);
		
		if(!$result){
			$this->logError($L['FileSystemFailedChmod'] . ' ' . $this->getChmodString() . ' ' . $L['FileSystemFailedChmodOnFile'] . ': ' . $path . '');
			return false;
		}
		return true;
	}
	
	/**
	 * logs the error in the database and sets the $error class variable
	 * 
	 * @param string $error the error message
	 */
	function setError($error) {
		$this->logError($error);
		$this->error = $error;
	}
	
	/**
	 * checks if there are any errors
	 * 
	 * @return boolean true if there is an error, else false
	 */
	function hasError() {
		if ($this->error) {
			return true;
		}
		return false;
	}
	
	/**
	 * returns the error message class variable
	 * 
	 * @return string the error message
	 */
	function printError() {
		return $this->error;
	}
	
	/**
	 * gets the contents of a directory and returns it as an array
	 *
	 * @param string $path the path of the directory
	 * @param boolean $recursive true to get sub directorys 
	 * @return array a multidimensional array for files and directories (with 
	 * the keys 'name' as the files name and 'dir' as a list sub directories and
	 * files)
	 */
	function getDirContents($path = false, $recursive = false, $exclusions = array()) {
		$exclusions = (array)$exclusions;
		$files = array();
		if (!$path) {
			$path = $this->getDirectory();
		}
		if (@file_exists($path) && @is_dir($path) && @is_readable($path)){
			if ($handle = opendir($path)) {
				while (false !== ($file = readdir($handle))) {
					if ( ($file != ".")  && ($file != "..") && (!in_array($file, $exclusions)) ) {
						$files[$path . '/' . $file]['name'] = $file;
						// if ( (filetype($path . '/' . $file) == 'dir') && $recursive) {
							// $files[$path . '/' . $file]['dir'] = $this->getDirContents($path . '/' . $file);
						// }
						if ( (@is_dir($path . '/' . $file)) && @is_readable($path . '/' . $file) && $recursive) {
							$files[$path . '/' . $file]['dir'] = $this->getDirContents($path . '/' . $file, $recursive, $exclusions);
						}
					}
				}
				closedir($handle);
			}
		}
		// $this->printDebug('getDirContents for ' . $path);
		// $this->printDebug($files);
		return $files;
	}
	
	/**
	 * Opens a file
	 * 
	 * @param string $file_location the complete location of the file
	 * @param string $rights rights to open the file with; see fopen
	 * @return boolean TRUE on success or FALSE on failure
	 */
	function openFile($file_location, $rights = 'r') {
		if ($this->current_file = @fopen($file_location,  $rights)) {
			return true;
		}
		global $SERVICES, $L;
		switch ($rights) {
			case 'w':
				$this->logError($L['FileSystemErrorCreatingFile'] . ': ' . $file_location);
			break;
			
			case 'a':
				$this->logError($L['FileSystemErrorAppendingToFile'] . ': ' . $file_location);
			break;
			
			case 'r':
				$this->logError($L['FileSystemErrorOpeningFile'] . ': ' . $file_location);
			break;
		}
		return false;
	}
	
	/**
	 * Closes a file the current file
	 */
	function closeFile() {
		// Phil 20060526 added check and @ (incase current file isnt a valid resource)
		// To be more pedantic we also could check get_resource_type($this->current_file) == 'file'
		if (is_resource($this->current_file)) {
			@fclose($this->current_file);
		}
	}
	
	/**
	 * Writes to a file
	 * 
	 * @param string $content the content that you want to write to the file
	 * @return boolean TRUE on success or FALSE on failure
	 */
	function writeToFile($content) {
		//MCR 20060419 changed this function around; using ===
		if (fwrite($this->current_file, $content) === false) {
			global $L;
			$this->setError($L['FileSystemErrorWritingToFile']);
			return false;
		}
		return true;
	}
	
	/**
	 * reads a file
	 *
	 * This is an internal function only. Should not be called anywhere except {@link readFileContents}
	 * 
	 * @param integer $readsize the number of bytes to read from the file
	 * @return mixed file contents on success or FALSE on failure
	 * @access private
	 */
	function getFileContents($readsize = false) {
		global $L;
		// Phil 20060526 New implementation
		// current_file has to either be a resource
		// Of course, not all resources are file handles, but this will have to do for now.
		// See readFileContents
		// To be more pedantic we could check get_resource_type($this->current_file) == 'file'
		if ($this->current_file && is_resource($this->current_file)) {
			// Phil 20060614 added @
			if ($file_content = @fread($this->current_file, $readsize)) { 
				return $file_content;
			} else {
				$this->setError($L['FileSystemErrorReadingFile']);
				return false;
			}
		} else {
			// Phil 20060526: Technically: 'Handle is not a valid resource'
			$this->setError($L['FileSystemErrorFileDoesNotExist']);
			return false;
		}
		return false;
	}
	
	/**
	 * reads the current line from a file and moves the pointer to the next line
	 * 
	 * @param integer $readsize the number of bytes to read from the file
	 * @return mixes file contents on success or FALSE on failure
	 */
	function getFileLine($readsize = false){
		if ($file_content = fgets($this->current_file, $readsize)) {
			return $file_content;
		}
		return false;
	}
	
	/**
	 * Reads the contents of a file
	 *
	 * @param int/string $id unique-per-module identifier of the file
	 * @param string $module module name
	 * @param boolean $return_as_array whether to return as a line_by_line array
	 * @param integer $readsize the number of bytes default is all
	 * @param mixed $explicit_path path to file
	 * @return string returns the contents of the file
	 */
	function readFileContents($id, $module, $return_as_array = false, $readsize = false, $explicit_path = false) {
		global $L;
		
		if ($explicit_path) {
			$file_location = $explicit_path; 
		} else {
			$file_location = $this->getFilePath($id, $module);
		}
		
		if (@file_exists($file_location)) {
			// Phil 20051021 - moved into this if block (file_exists) from above
			if (!is_readable($file_location)){
				$this->setError($L['FileSystemErrorReadingFile'] . ': ' . $file_location);
				return false;
			}
			if ($return_as_array) {
				return file($file_location);
			}
			if (!$readsize) {
				$readsize = filesize($file_location);
			}
			if (!$this->openFile($file_location, 'r')) {
				return false;
			}
			$file_content = $this->getFileContents($readsize);
			$this->closeFile();
			return $file_content;
			
		} else {
			$this->setError($L['FileSystemErrorFileDoesNotExist'] . ': ' . $file_location);
			return false;
		}
	}
	
	/**
	 * Writes a file to disk
	 * 
	 * @param int/string $id unique-per-module identifier of the file
	 * @param string $module module name
	 * @param string $content the content for the file
	 * @param boolean $overwrite true to overwrite otherwise false
	 * @param mixed $explicit_path path to file
	 * @return boolean TRUE on success or FALSE on failure
	*/
	function writeFileContents($id, $module, $content, $overwrite = true, $explicit_path = false){
		global $SERVICES;
		if ($explicit_path) {
			$file_location = $explicit_path; 
		} else {
			$file_location = $this->getFilePath($id, $module);
		}
		
		if (@file_exists($file_location)) {
			if($overwrite == true){
				if (!$this->openFile($file_location, 'w')) {
					return false;
				}
			} else {
				$this->setError($L['FileSystemErrorFileAlreadyExists'] . ': ' . $file_location);
	   			$this->closeFile();
				return false;
			}
		} else {
			if (!$this->openFile($file_location, 'w')) {
				return false;
			}
			$this->doChmod($file_location);
		}
		if(!$this->writeToFile($content)){
			$this->closeFile();
			return false;
		}
		return true;
	}

	/**
	 * gets the MIME type for a file if the file dosen't exist or is a 
	 * directory returns false.
	 * 
	 * @param string $filename the file name
	 * @param mixed $path to overwirte the directoy form the clas var
	 * @return mixed the MIME type or false
	 */
	function getFileMimeType($filename, $path = false) {
		if (!$path) {
			$path = $this->getDirectory();
		}
		if (!function_exists('mime_content_type')) {
			function mime_content_type($f) {
				$f = escapeshellarg($f);
				return trim( `file -bi $f` );
			}
		}
		if (@filetype($path . '/' . $filename) == 'file') {
			return mime_content_type($path . '/' . $filename);
		}
		return false;
	}
	
	/**
	 * gets the size for a file 
	 * if the file dosen't exist or is a directory, returns false.
	 * 
	 * @param int/string $id identifier
	 * @param string $module module name
	 * @return mixed the file size or false
	 */
	function getFileSize($id, $module) {
		$file_location = $this->getFilePath($id, $module);
		
		if (@file_exists($file_location)){
			return filesize($file_location);
		} else {
			$this->setError($L['FileSystemErrorFileDoesNotExist'] . ': ' . $file_location);
			return false;
		}
	}
	
	/**
	 * checks if the file exists
	 * 
	 * @param string $filename the file name
	 * @param mixed $path to overwirte the directoy form the clas var
	 * @return boolean true if file exists, false otherwise
	 */
	function checkFileExists($filename, $path = false) {
		if (!$path) {
			$path = $this->getDirectory();
		}
		return @file_exists($path . '/' . $filename);
	}
	
	/**
	 * checks if the directory exists
	 * 
	 * @param string $filename the file name
	 * @param mixed $path to overwirte the directoy form the clas var
	 * @return boolean true if file exists, false otherwise
	 */
	function checkDirExists($dirname, $path = false) {
		// print '<div>FileSystem::checkDirExists(' . $dirname . ', ' . $path . ')</div>';
		if (!$path) {
			$path = $this->getDirectory();
		}
		return ( @file_exists($path . '/' . $dirname) && @is_dir($path . '/' . $dirname) );
	}
	
	/**
	 * prepare a file for open/save in browser
	 * 
	 * @param integer $id the item_id
	 * @param string $module the module name
	 * @param string $master_dir the top level directory in the data directory (optional)
	 * @param boolean TRUE to open the file as a download or FALSE to view in the browser
	 */
	function getFileForBrowser($id, $module, $master_dir = false, $sub_dir = false, $file_info = false, $download = false) {
		$path = $this->getFilePath($id, $module, $master_dir, $sub_dir);
		//TODO don't think master_dir, sub_dir are used. should remove 
		$this->setDirectory(dirname($path));
		// print '<div>1) DEBUG FileSystem, id: ' . $id .', module: ' . $module . ', master_dir: ' . $master_dir . ', sub_dir: ' . $sub_dir . ', download: ' . $download . ', file_info:</div>'; print_r($file_info); print '<hr>'; 
		if (!$file_info) {
			$file_info = array();
			$file_info['mime_type'] = $this->getFileMimeType(basename($path));
			$file_info['file_name'] = basename($path);
		}
		if (!$file_info['location']) {
			$file_info['location'] = $path;
		}
		if (!$file_info['file_name']) {
			$file_info['file_name'] = $id;
		}
		
		if (!$file_info['real_file_name']) {
			$file_info['real_file_name'] = $id;
		}
		// print '<div>DEBUG FileSystem, id: ' . $id .', module: ' . $module . ', master_dir: ' . $master_dir . ', sub_dir: ' . $sub_dir . ', download: ' . $download . ', file_info:</div>'; print_r($file_info); print '<hr>'; 
		// print '<div>DEBUG FileSystem: ' . $v->d_parameters['filename'] . '</div>'; print_r($v->d_parameters); print '<hr>'; exit();
		return $this->displayFile($file_info, $download);
	}
	
	/**
	 * displays the file as a 'view' or 'download' in the browser
	 * 
	 * @param array $file_info the file details; 
	 *   required array elements are 'file_name', 'mime_type' and either 'location' or 'content'
	 * @param boolean TRUE to open the file as a download or FALSE to view in the browser
	 * @return filestream the file as either a download or to view in the browser
	 */
	function displayFile($file_info, $download = false) {
		global $L;
		// switch here to cater for the 2 different ways to call this function
		// either file pointer, or file content
		if ($file_info['location']) {
			if (!@file_exists($file_info['location']) || !@is_readable($file_info['location'])) {
				$this->setError($L['FileSystemErrorFileDoesNotExist'] . ': ' . $file_info['file_name']);
				return false;
			}
		}
		
		//echo 'download '; print($download);echo 'file '; print_r($file_info);
		//2005-05-02 MCR   header('Content-type: '. $file_info['mime_type']);
		if ($download) {
			//global $SERVICES;
			//$user_agent = $SERVICES->getGlobal('user_agent');
			
			//2005-05-02 MCR   
			//this block is from our previous export module, with heritage from pma
			header('Content-Type: application/x-download');
			header('Content-Disposition: attachment; filename="' . $file_info['file_name'] . '"');
			header('Expires: 0');
			header('Cache-Control: must-revalidate, post-check=0, pre-check=0');
			header('Pragma: public');
			
		} else {  //attempt to view in browser (provided browser knows extension) , 2005-05-02 MCR
			
			header('Content-Type: ' . $file_info['mime_type']);
			header('Content-Disposition: inline; filename="' . $file_info['file_name'] . '"');   //this works in IE6, MCR
					//see http://support.microsoft.com/default.aspx?scid=KB;en-us;q260519
					
			header('Expires: 0');
			header('Cache-Control: must-revalidate, post-check=0, pre-check=0');
			header('Pragma: public');
			
		}
		
		
		// $this->printDebug('before, file_info:'); $this->printDebug($file_info);
		
		//2005-05-02 MCR   header("Content-Length: " . filesize($file_info['location']) );
		// switch here to cater for the 2 different ways to call this function
		// either file pointer, or file content
		if ($file_info['location']) {
			// if ($file = fopen($file_info['location'], 'rb')) {		
				// fpassthru($file);
				// fclose($file);
			// }
			
			// Phil 20060707 - Enables downloading files > the php memory limit
			if ($file = fopen($file_info['location'], 'rb')) {
				// $this->printDebug('before, file_info:');
				while ( (!feof($file)) && (connection_status()==0) ) {
					print(fread($file, 1024*8));
					// $this->printDebug('...');
					ob_flush();
					flush();
				}
				fclose($file);
			}
			// Never used, possible future enhancement // return((connection_status()==0) and !connection_aborted());
			// $this->printDebug('...after');
			
		} else if ($file_info['content']) {
			echo $file_info['content'];
		}
	}
	
	/**
	 * gets a files last modified information
	 * 
	 * @param string $filename the file name
	 * @param mixed $path to overwirte the directoy form the clas var
	 * @return UNIX TIMESTAMP
	 */
	function getFileLastModified($filename, $path = false) {
		if (!$path) {
			$path = $this->getDirectory();
		}
		return filemtime($path . '/' . $filename);
	}
	
	/**
	 * deletes a file from the webserver
	 * 
	 * @param string $filename the file name
	 * @param mixed $path to overwirte the directoy form the clas var
	 * @return boolean TRUE on success or FALSE on failure
	 */
	function deleteFile($filename, $path = false) {
		if (!$path) {
			$path = $this->getDirectory();
		}
		if (@file_exists($path . '/' . $filename)) {
			return @unlink($path . '/' . $filename);
		}
		return false;
	}
	
	/**
	 * removes a file from the temp directory
	 * 
	 * @param string the file name
	 * @return boolean TRUE on success or FALSE on failure
	 */
	function collectGarbage($filename) {
		global $SERVICES;
		$SERVICES->getService('Config');
		$tmp_path = $SERVICES->config->lookup('files.tmpdir');
		return $this->deleteFile($filename, $tmp_path);
	}
	
	/**
	 * handles a file upload
	 *
	 * @param array $file_item the name of the form item
	 * @param boolean $save_file true if you want to save the file, or  false 
	 * to get the contents
	 * @return array an array of file information (with keys 'name', 'type', 
	 * 'error' and 'contents' or 'location')
	 */
	function processUpload($file_item, $save_file = false) {
		$file_info = array();
		// $this->printDebug('processUpload(' . $file_item . '' . ($save_file ? 'true' : 'false'));
		// $this->printDebug('processUpload $_FILES:');
		// $this->printDebug($_FILES);
		//	print_r($_FILES['item']['size'][$file_item]);
		if ($_FILES['item']['name'][$file_item]) {
			if ($save_file) {
				global $SERVICES, $L;
				$SERVICES->getService('Config');
				// Phil 20060623 Ensure the bcrm tmp dir exists
				$bcrm_tmpdir = $SERVICES->config->lookup('files.tmpdir');
				if ($bcrm_tmpdir && (!@file_exists($bcrm_tmpdir) || !@is_writable($bcrm_tmpdir))) {
					$dir_name = basename($bcrm_tmpdir);
					$path = dirname($bcrm_tmpdir);
					$this->createDir($dir_name, $path, true);
				}
				
				$save_path = tempnam($bcrm_tmpdir, $_FILES['item']['name'][$file_item] . '.');
				//$save_path = $_FILES['item']['tmp_name'][$file_item];
				// $this->printDebug(' save_path is: ' . $save_path);
				//echo ' save_path is: ' . $save_path;
				
				// check if it's the uploaded file (protects against hacking system files)
				if (!is_uploaded_file($_FILES['item']['tmp_name'][$file_item])) {
					$this->setError($L['FileSystemErrorFileNotUploaded'] .' (' . $_FILES['item']['error'][$file_item] .')');
				} else if (!is_readable($_FILES['item']['tmp_name'][$file_item]) || !file_exists($_FILES['item']['tmp_name'][$file_item])) {
					// check if the file exists & it's readable
					$this->setError($L['FileSystemErrorNotReadable'] .' (' . $_FILES['item']['error'][$file_item] .')');
				} else if (@move_uploaded_file($_FILES['item']['tmp_name'][$file_item], $save_path)) {
					// finally, try to move it
					// echo "<br>----here we are, upload done<br>";
					$file_info['location'] =  $save_path;
				} else {
				//	echo "<br>-----------Error has been set------<br>";
					$this->setError($L['FileSystemErrorMoveFailed'] .' (' . $_FILES['item']['error'][$file_item] .')');
				}
			} else {
				// TODO: Phil 20060704 This block will fail. Called by Output if 'get_content' parameter specified
				// filename not specified (commented out), 
				// readFileContents is a class method (needs $this->)
				// Have corrected these errors but not tested
				
				//$explicit_path = rtrim(dirname($_FILES['item']['tmp_name'][$file_item]), '/');
				//$filename = str_replace($path . '/', '', $_FILES['item']['tmp_name'][$file_item]);
				$explicit_path = $_FILES['item']['tmp_name'][$file_item];
				
				//TODO test; echo '**' . $explicit_path;
				//$file_info['contents'] = $this->readFileContents($filename, false, $path);
				$filename = false; // Phil 20060704
				$file_info['contents'] = $this->readFileContents($filename, 'tmp', false, false, $explicit_path); //only $explicit_path is used here
				if ($_FILES['item']['error'][$file_item]) {
					$this->setError($_FILES['item']['error'][$file_item]);
				}
			}
			$file_info['name'] = $_FILES['item']['name'][$file_item];
			$file_info['type'] = $_FILES['item']['type'][$file_item];
			$file_info['error'] = $_FILES['item']['error'][$file_item];
			$file_info['size'] = $_FILES['item']['size'][$file_item];
			//print_r($file_info);
			// $this->printDebug('processUpload(' . $file_item . '' . ($save_file ? 'true' : 'false') . ' returning');
			return $file_info;
		}
	}
	
	/**
	 * creates a directory
	 * 
	 * @param string $new_dir the new directory name
	 * @param string $path the base path for the directory (optional)
	 * @param boolean $ignore_exists true to not display an error if already 
	 * exists else false (optional)
	 * @return boolean TRUE on success or FALSE on failure
	 */
	function createDir($new_dir, $path = false, $ignore_exists = false) {
		global $L;
		if (!$path) {
			$path = $this->getDirectory();
		}
		if(!@is_dir($path . '/' . $new_dir)) {
			if(!@mkdir($path . '/' . $new_dir, $this->getChmodString())) {  
				$this->setError($L['FileSystemErrorCreatingDir'] . ': ' . $path . '/' . $new_dir);
				return false;
			}
			if (!$this->doChmod($path . '/' . $new_dir)) {
				$this->setError($L['FileSystemErrorChmodDir'] . ': ' . $path . '/' . $new_dir);
				return false;
			}
			
		} else {
			if (!$ignore_exists) {
				$this->setError($L['FileSystemDirAlreadyExists'] . ': ' . $path . '/' . $new_dir);
				return false;
			}
		}
		return true;
	}
	
	/**
	 * copies a file on the server
	 * 
	 * @param string $source the source file including the full path
	 * @param srting $new_name the new file name (optional)
	 * @param string $path the path to copy the file to (optional)
	 * @return boolean TRUE on success or FALSE on failure
	 */
	function copyFile($source, $new_name = false, $path = false) {
		//echo $source . ' -------------- ' . $new_name . ' -------- ' . $path . '****';
		if (!$path) {
			$path = $this->getDirectory();
		}
		if ($new_name) {
			$dest = $path . '/' . $new_name;
		} else {
			$dest = $path . '/' . basename($source);
		
		}
		
		if (!copy($source, $dest)) {
			global $L;
			$this->setError($L['FileSystemErrorCopyingFile'] . ' [copyFile]: ' . $source . ' >> ' . $dest);
			return false;
		} else {
			$this->doChmod($dest);
		}
		return true;
	}	
	
	/**
	 * recursively copies a directory
	 * 
	 * @param string $source the path to the source directory
	 * @param string $path the path of the destination directory (optional)
	 * @param array $exclusions An optional array of direcories (full path) to exclude
	 * @return boolean TRUE on success or FALSE on failure
	 */
	function recursiveCopy($source, $path = false, $exclusions = array()) {
		if (!$path) {
			$path = $this->getDirectory();
		}
		
		$result = true;
		$dir_contents = $this->getDirContents($source, true);
		
		$source = rtrim($source, '/');
		$new_dir = substr ($source, strrpos($source, '/') + 1);
		$destination = $path . '/' . $new_dir;
		
		// $this->printDebug(get_class($this) . '::recursiveCopy , path: ' . $path . ', source: ' . $source . ', destination: ' . $destination);
		
		
		if (!@file_exists($destination) || !@is_dir($destination)) {
			// $this->printDebug(get_class($this) . '::recursiveCopy desination needs to be created: ' . $destination);
			if (!$this->createDir($new_dir, $path)) {
				// print '<p>PROBLEM createDir: ' . $new_dir . ' in ' . $path . '</p>';
				// $this->printDebug('ERROR: ' . get_class($this) . '::recursiveCopy failed to create directory ' . $destination);
				return false;
			}
		}
		
		foreach ($dir_contents as $k => $v) {
			$result = true;
			if ($k && !in_array($k, $exclusions)) {
				// print '<div>include: ' . $k . '</div>'; print_r($v); print '<hr>';
				if (isset($v['dir'])) { // Phil: can be empty
					// $this->printDebug('[dir] ' . get_class($this) . '::recursiveCopy calling recursiveCopy ' . $k . ' to ' . $destination);
					$result = $this->recursiveCopy($k, $destination);
				} else {
					// $this->printDebug('[file] ' . get_class($this) . '::recursiveCopy calling recursiveCopy ' . $k . ' to ' . $destination);
					$result = $this->copyFile($k, false, $destination);
				}
				
				if ($result === false) {
					// $this->printDebug('ERROR: ' . get_class($this) . '::recursiveCopy failed to copy ' . $k . ' to ' . $destination);
					return $result;
				}
			} // else {
				// $this->printDebug('Skipping: ' . $k); $this->printDebug($v);
			// }
		}
		
		// if (!$result) {
			// $this->printDebug('ERROR: ' . get_class($this) . '::recursiveCopy failed');
		// }
		return ($result ? true : false);
	}
	
	
	/**
	 * moves a file to the correct data folder
	 *
	 * Parameters $master_dir and $src_dir deprecated version 4.6
	 * 
	 * @param integer $id the item_id
	 * @param string $module the module name
	 * @param string $source the location of the source file (must be the full path)
	 * @param string $sub_dir the subdirectory to store the file in
	 * @return boolean TRUE on success or FALSE on failure
	 */
	function moveFile($id, $module, $source) {
		global $SERVICES, $L;
		$dest = $this->getFilePath($id, $module);
		
		//copy the file; 
		//not this->copyFile function as we are writing to a data slot, not a dir location
		if (!copy($source, $dest)) {
			$this->setError($L['FileSystemErrorCopyingFile'] . ' [moveFile]: ' . $source . ' >> ' . $dest);
			return false;
		} else {
			$this->doChmod($dest);
		}
		
		//delete source file
		$source_dir = dirname($source);
		if (!$this->setDirectory($source_dir)) {
			return false;
		}
		if (!$this->deleteFile(basename($source))) {
			return false;
		}
		return true;
	}
	
	/**
	 * returns the path and file name of a file
	 * 
	 * @param integer $id the item_id
	 * @param string $module the module name
	 * @param string $master_dir the top level directory in the data directory (optional)
	 * @return string the file path
	 */
	function getFilePath($id, $module, $master_dir = false, $sub_dir = false) {
		global $SERVICES;
		$SERVICES->getService('Config');
		
		//TODO don't think master_dir, sub_dir are used. should remove
		if ($master_dir || $sub_dir) {
			// echo '<h1> master or sub dir set';
		}
		
		$files_dir = $SERVICES->config->lookup('files.dir');
		
		if ($id && is_numeric($id)) {
			
			$item_dir = floor($id / $this->max_files_per_dir);
			$file = $id;
			
			$path .= $files_dir . '/' . $module . '/' . $item_dir;
			
			if (!@file_exists($path . '/' . $file)) {
				
				if (!$this->setDirectory($files_dir)) {
					return false;
				}
				if (!$this->createDir($module, false, true)) {
					return false;
				}
				if (!$this->setDirectory($files_dir . '/' . $module)) {
					return false;
				}		
				if (!$this->createDir($item_dir, false, true)) {
					return false;
				}
				if (!$this->setDirectory($files_dir . '/' . $module . '/' . $item_dir)) {
					return false;
				}
			}
			
			if (!$this->setDirectory($path)) {
				return false;
			}
			
		} else {
			//this is used by AdminTools [only]
			//TODO, update function to pass param instead of testing for non numeric
			$file = $id;
		}
		return $path . '/' . $file;
	}
	
	/**
	 * deletes an empty directory from the webserver
	 * 
	 * @param string $directory the directory name
	 * @return boolean TRUE on success or FALSE on failure
	 * @author Phil Calder
	 * @access private
	 */
	function deleteDirectory($directory) {
		if (@file_exists($directory) && @is_dir($directory)) {
			// $this->printDebug('deleting directory: ' . $directory);
			return @rmdir($directory);
		}
		// $this->printDebug('Cannot delete directory: ' . $directory . ' (doesnt exist or is not a directory)');
		return false;
	}
	
	/**
	 * recursively deletes a directory and all files in it
	 *
	 * The directory must be in the applications backup path or cache directory
	 * 
	 * @param string $source the path to the source directory
	 * @param array $exclusions An optional array of direcories (full path) to exclude
	 * @return boolean TRUE on success or FALSE on failure
	 * @author Phil Calder
	 * @access private
	 */
	function recursiveDelete($path, $exclusions = array()) {
		global $SERVICES, $BCRM_DOCUMENT_ROOT;
		$result = true;
		$backup_path = $SERVICES->config->lookup('modules.Admin.backup.dir');
		$cache_dir =  $BCRM_DOCUMENT_ROOT . '/cache';
		if (!$backup_path && !$cache_dir) {
			// $this->printDebug('Error: backup path and cache dir is not set [' . $backup_path . ']');
			return false;	
		}
		$not_backup_dir = (substr($path, 0, strlen($backup_path)) != $backup_path);
		$not_cache_dir = (substr($path, 0, strlen($cache_dir)) != $cache_dir);
		
		if ($not_backup_dir && $not_cache_dir) {
			// $this->printDebug('Error: directory to delete [' . $path . '] must be in the backup path [' . $backup_path . '] or cache dir [' . $cache_dir . '] ');
			return false;
		}
		// print '<div>deleting dir: ' . $path . '</div>'; exit();
		$dir_contents = $this->getDirContents($path, true);
		
		$source = rtrim($path, '/');
		
		foreach ($dir_contents as $k => $v) {
			$result = true;
			if ($k && !in_array($k, $exclusions)) {
				// $this->printDebug('recursiveDelete k=: ' . $k);
				// $this->printDebug($v);
				
				if (isset($v['dir'])) {
					// $this->printDebug('calling recursiveDelete: ' . $k); // $this->printDebug($v);
					
					$result = $this->recursiveDelete($k, $exclusions);
					// if (count($v['dir'])) {
						// $this->printDebug('recursiveDelete for: ' . $k . ' returned: ' . ($result ? 'true' : 'false') );
					// } 
					// if ($result !== false) {
						// $this->printDebug('calling deleteDirectory: ' . $k);
						// $result = $this->deleteDirectory($k);
					// }
				} else {
					// $this->printDebug('calling deleteFile(' . basename($k) . ', ' . dirname($k) . ')');
					$this->deleteFile(basename($k), dirname($k));
				}
				
				if ($result === false) {
					// $this->printDebug('recursiveDelete, problem deleting: ' . $k); // $this->printDebug($v);
					// print '<p>PROBLEM with: ' . $k . ' in ' . $destination . '</p>';
					return $result;
				}
			} // else {
				// print '<div>exclude: ' . $k . '</div>'; print_r($v); print '<hr>';
			// }
		}
		
		if ( ($result !== false) && @file_exists($path) && @is_dir($path)) {
			$result = $this->deleteDirectory($path);
			// $this->printDebug('deleteDirectory ' . $path . ' returned ' . ($result ? 'true' : 'false'));
		} // else {
			// $this->printDebug('Directory ' . $path . ' cannot be deleted');	
		// }
		return ( ($result !== true) ? false : true);
	}
	
	/**
	 * Zips a directory
	 * 
	 * The archived file will be placed in the applications backup dir
	 *
	 * @param string $output_filename the name for the archive, eg 'mybackup'
	 * @param string $path the path to directory that you want to zip (optional)
	 * @param boolean $is_relative_to_backup_dir the parameter path is a sub directory of the application backup directory (optional)
	 * @return boolean TRUE on success or FALSE on failure
	 * @access private
	 * @author Phil Calder
	 */
	function zipDirectory($output_filename, $path = false, $is_relative_to_backup_dir = false) {
		// $this->printDebug(get_class($this) . '::zipDirectory(' . $output_filename . ', ' . $path . ') [begin]');
		global $SERVICES;
		$backup_path = $SERVICES->config->lookup('modules.Admin.backup.dir');
		
		if (!$path) {
			$path = $this->getDirectory();
		} else {
			if ($is_relative_to_backup_dir && $backup_path) { // Specified path arg is a dir name in the backup directory
				$path = $backup_path . '/' . $path; 
			}
		}
		
		$type = 'tar.gz'; // Use tar.gz, zip not tested
		
		// Can use Archive_Tar or Archive_Zip
		$package_name = 'archive';
		
		if ($type == 'tar.gz') {
			$file_name = 'Tar.php';
			$class_name = 'Archive_Tar';
			$compression_type = 'gz';
			$file_ext = '.tar.gz';
		} else {
			$file_name = 'Zip.php';
			$class_name = 'Archive_Zip';
			$compression_type = 'zip';
			$file_ext = '.zip';
		}
		
		// $this->printDebug(get_class($this) . '::zipDirectory calling includeExternal(' . $package_name . ', ' . $file_name . ')');
		if (!includeExternal($package_name, $file_name)) {
			// $this->printDebug('Cannot load external package: ' . $package_name . ' or file ' . $file_name);
			return false;
		}
		
		$dir_to_backup = $path;
		if ($dir_to_backup && @file_exists($dir_to_backup) && @is_dir($dir_to_backup) && @is_readable($dir_to_backup)) {
			// $this->printDebug('Dir to backup: "' . $dir_to_backup . '" exists');
			if (chdir(dirname($dir_to_backup))) {
				// $this->printDebug('Changed working dir: ' . dirname($dir_to_backup));
				$dir_to_backup = basename($dir_to_backup);
				// $this->printDebug('Relative dir to backup: ' . $dir_to_backup);
			} // else {
				// $this->printDebug('Dir to backup: "' . $dir_to_backup . '" exists in  ' . getcwd());
			// }
		} else {
			// Cannot find/use directory to backup
			// $this->printDebug('Cannot find or use directory to backup: "' . $dir_to_backup . '"');
			return false;
		}
		if ($dir_to_backup && @file_exists($dir_to_backup) && @is_dir($dir_to_backup) && @is_readable($dir_to_backup)) {
			// $this->printDebug('Dir to backup: "' . $dir_to_backup . '" exists in  ' . getcwd());
		} else {
			// Cannot find/use directory in - current working dir - to backup
			// $this->printDebug('Dir to backup: "' . $dir_to_backup . '" DOES NOT exist in  ' . getcwd());
			return false;
		}
		
		$backup_path = $SERVICES->config->lookup('modules.Admin.backup.dir');
		if ($backup_path && @file_exists($backup_path) && @is_writable($backup_path)) {
			// $this->printDebug('Backup path is available: "' . $backup_path . '"');
		} else {
			// $this->printDebug('Backup path is NOT available: "' . $backup_path . '"');
			return false;
		}
		
		// $this->printDebug('Full path: ' . $path);
		// $this->printDebug('Working dir: ' . getcwd());
		// $this->printDebug('Dir to backup: ' . $dir_to_backup);
		
		$output_filename = trim($output_filename);
		// TODO: rewrite these statements (so they work correctly)
		// $output_filename = rtrim($output_filename, '.gz');
		// $output_filename = rtrim($output_filename, '.tar');
		if (substr($output_filename, -(strlen($file_ext))) == $file_ext) {
			$output_filename = substr($output_filename, 0, (strlen($output_filename) - strlen($file_ext)));
		}
		
		// $output_filename = $backup_path . '/' . $output_filename . '.tar.gz';
		$output_filename = $backup_path . '/' . $output_filename . $file_ext;
		// $this->printDebug('Output filename: "' . $output_filename . '"');
		// $this->printDebug('Checking whether class exists ' . $class_name . '...');
		if (class_exists($class_name)) {
			// $this->printDebug('Class exists, creating instance...');
			$compression_class = new $class_name($output_filename, $compression_type);
			// $this->printDebug('Instance created');
			// Create a list of files and directories
			$list = array($dir_to_backup);
			
			// Create the zip file
			if ($compression_class->create($list)) {
				// $this->printDebug('Output file created ' . $output_filename);
			} else {
				// $this->printDebug('Output file NOT created ' . $output_filename);
				return false;
			}
		} else {
			// $this->printDebug('Compression class does not exist: ' . $class_name);
			return false;
		}
		
		return true;
	}
} // class FileSystem
 
?>